// Copyright (c) 2022 The Bitcoin Core developers
// Distributed under the MIT software license, see the accompanying
// file COPYING or http://www.opensource.org/licenses/mit-license.php.

#include <chainparams.h>
#include <node/blockstorage.h>
#include <node/context.h>
#include <validation.h>

#include <boost/test/unit_test.hpp>
#include <test/util/setup_common.h>

using node::BLOCK_SERIALIZATION_HEADER_SIZE;
using node::BlockManager;
using node::MAX_BLOCKFILE_SIZE;

// use BasicTestingSetup here for the data directory configuration, setup, and
// cleanup
BOOST_FIXTURE_TEST_SUITE(blockmanager_tests, BasicTestingSetup)

BOOST_AUTO_TEST_CASE(blockmanager_find_block_pos) {
    const auto params{CreateChainParams(*m_node.args, CBaseChainParams::MAIN)};
    const BlockManager::Options blockman_opts{
        .chainparams = *params,
        .blocks_dir = m_args.GetBlocksDirPath(),
    };
    BlockManager blockman{blockman_opts};
    CChain chain{};
    // simulate adding a genesis block normally
    BOOST_CHECK_EQUAL(
        blockman.SaveBlockToDisk(params->GenesisBlock(), 0, chain, nullptr)
            .nPos,
        BLOCK_SERIALIZATION_HEADER_SIZE);
    // simulate what happens during reindex
    // simulate a well-formed genesis block being found at offset 8 in the
    // blk00000.dat file the block is found at offset 8 because there is an 8
    // byte serialization header consisting of 4 magic bytes + 4 length bytes
    // before each block in a well-formed blk file.
    FlatFilePos pos{0, BLOCK_SERIALIZATION_HEADER_SIZE};
    BOOST_CHECK_EQUAL(
        blockman.SaveBlockToDisk(params->GenesisBlock(), 0, chain, &pos).nPos,
        BLOCK_SERIALIZATION_HEADER_SIZE);
    // now simulate what happens after reindex for the first new block processed
    // the actual block contents don't matter, just that it's a block.
    // verify that the write position is at offset 0x12d.
    // this is a check to make sure that
    // https://github.com/bitcoin/bitcoin/issues/21379 does not recur 8 bytes
    // (for serialization header) + 285 (for serialized genesis block) = 293 add
    // another 8 bytes for the second block's serialization header and we get
    // 293 + 8 = 301
    FlatFilePos actual{
        blockman.SaveBlockToDisk(params->GenesisBlock(), 1, chain, nullptr)};
    BOOST_CHECK_EQUAL(
        actual.nPos,
        BLOCK_SERIALIZATION_HEADER_SIZE +
            ::GetSerializeSize(params->GenesisBlock(), CLIENT_VERSION) +
            BLOCK_SERIALIZATION_HEADER_SIZE);
}

BOOST_FIXTURE_TEST_CASE(blockmanager_scan_unlink_already_pruned_files,
                        TestChain100Setup) {
    // Cap last block file size, and mine new block in a new block file.
    const auto &chainman = Assert(m_node.chainman);
    auto &blockman = chainman->m_blockman;
    const CBlockIndex *old_tip{
        WITH_LOCK(chainman->GetMutex(), return chainman->ActiveChain().Tip())};
    WITH_LOCK(chainman->GetMutex(),
              blockman.GetBlockFileInfo(old_tip->GetBlockPos().nFile)->nSize =
                  MAX_BLOCKFILE_SIZE);
    CreateAndProcessBlock({}, GetScriptForRawPubKey(coinbaseKey.GetPubKey()));

    // Prune the older block file, but don't unlink it
    int file_number;
    {
        LOCK(chainman->GetMutex());
        file_number = old_tip->GetBlockPos().nFile;
        blockman.PruneOneBlockFile(file_number);
    }

    const FlatFilePos pos(file_number, 0);

    // Check that the file is not unlinked after ScanAndUnlinkAlreadyPrunedFiles
    // if m_have_pruned is not yet set
    WITH_LOCK(chainman->GetMutex(), blockman.ScanAndUnlinkAlreadyPrunedFiles());
    BOOST_CHECK(!AutoFile(blockman.OpenBlockFile(pos, true)).IsNull());

    // Check that the file is unlinked after ScanAndUnlinkAlreadyPrunedFiles
    // once m_have_pruned is set
    blockman.m_have_pruned = true;
    WITH_LOCK(chainman->GetMutex(), blockman.ScanAndUnlinkAlreadyPrunedFiles());
    BOOST_CHECK(AutoFile(blockman.OpenBlockFile(pos, true)).IsNull());

    // Check that calling with already pruned files doesn't cause an error
    WITH_LOCK(chainman->GetMutex(), blockman.ScanAndUnlinkAlreadyPrunedFiles());

    // Check that the new tip file has not been removed
    const CBlockIndex *new_tip{
        WITH_LOCK(chainman->GetMutex(), return chainman->ActiveChain().Tip())};
    BOOST_CHECK_NE(old_tip, new_tip);
    const int new_file_number{
        WITH_LOCK(chainman->GetMutex(), return new_tip->GetBlockPos().nFile)};
    const FlatFilePos new_pos(new_file_number, 0);
    BOOST_CHECK(!AutoFile(blockman.OpenBlockFile(new_pos, true)).IsNull());
}

BOOST_AUTO_TEST_SUITE_END()
